"""
Extensions for aioclock.

AioClock is very extensible, and you can add your own extensions to it.
The extension would allow you to interact with your AioClock instance, from different layers of your application.
For instance, the FastAPI plugin allows you to run a specific task immediately from an HTTP API, or see your tasks in an HTTP API, and when they are going to run next.
"""
